""" Xibabel neuroimaging interface

Xibabel is an experimental package for working with neuroimaging data formats.
"""

from .loaders import load, save, from_array, to_nifti  # noqa:F401

__version__ = "0.0.1rc1"
